"""This is the demo script to show how to import data from OlapCubes and
Reports. It is possible to select attributes, metrics and attribute forms which
is presented.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.project_objects import OlapCube, Report

# get connection to an environment
base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
connection = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                        login_mode=1)

cube_id = "some_cube_id"
report_id = "some_report_id"

# get cube based on its id and store it in data frame
my_cube = OlapCube(connection=connection, id=cube_id)
my_cube_df = my_cube.to_dataframe

# get report based on its id and store it in data frame
my_report = Report(connection=connection, id=report_id, parallel=False)
my_report_df = my_report.to_dataframe

# get list of ids of metrics, attributes or attribue elements available within
# Cube or Report
my_cube.metrics
my_cube.attributes
my_cube.attr_elements

# by default all elements are shown in the data frame. To choose elements you
# have to pass proper IDs to function 'apply_filters()' which is available for
# Cube and Report
my_cube.apply_filters(
    attributes=["A598372E11E9910D1CBF0080EFD54D63", "A59855D811E9910D1CC50080EFD54D63"],
    metrics=["B4054F5411E9910D672E0080EFC5AE5B"],
    attr_elements=[
        "A598372E11E9910D1CBF0080EFD54D63:Los Angeles", "A598372E11E9910D1CBF0080EFD54D63:Seattle"
    ],
)
# check selected elements which will be placed into a dataframe
my_cube.selected_attributes
my_cube.selected_metrics
my_cube.selected_attr_elements

my_cube_applied_filters_df = my_cube.to_dataframe

# to exclude specific attribue elements, pass the `operator="NotIn"` to
# `apply_filters()` method
my_cube.apply_filters(
    attributes=["A598372E11E9910D1CBF0080EFD54D63", "A59855D811E9910D1CC50080EFD54D63"],
    metrics=["B4054F5411E9910D672E0080EFC5AE5B"],
    attr_elements=[
        "A598372E11E9910D1CBF0080EFD54D63:Los Angeles", "A598372E11E9910D1CBF0080EFD54D63:Seattle"
    ],
    operator="NotIn",
)

my_cube_exceluded_elements_df = my_cube.to_dataframe



from enum import Enum
from typing import Any, Dict, List, Optional, TYPE_CHECKING

from mstrio.access_and_security.security_filter import ObjectReference
from mstrio.utils.helper import Dictable

if TYPE_CHECKING:
    from mstrio.connection import Connection


class ParameterType(Enum):
    CONSTANT = "constant"
    OBJECT_REFERENCE = "object_reference"
    EXPRESSION = "expression"
    PROMPT = "prompt"
    ARRAY = "array"
    DYNAMIC_DATE_TIME = "dynamic_date_time"


class ParameterBase(Dictable):

    TYPE = None

    def __init__(self):
        self.parameter_type = self.TYPE


class ConstantType:
    INT_32 = "int32"
    INT_64 = "int64"
    DOUBLE = "double"
    DATE = "date",
    TIME = "time",
    STRING = "string"


class ConstantParameter(ParameterBase):

    TYPE = ParameterType.CONSTANT

    def __init__(self, type: "ConstantType", value: str):
        """Constant parameter for predicate.

        Args:
            type (enum): type of constant. Possible values are defined in enum
                `ConstantType`.
            value (str): value of constant.
        """
        super().__init__()
        self.constant = {"type": type, "value": value}

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source["type"] = source.get('constant', {}).get('type')
        source["value"] = source.get('constant', {}).get('value')
        return super().from_dict(source, connection)


class ObjectReferenceParameter(ParameterBase):

    TYPE = ParameterType.OBJECT_REFERENCE

    def __init__(self, target: "ObjectReference"):
        super().__init__()
        self.target = target

    _FROM_DICT_MAP = {"target": ObjectReference.from_dict}


class ExpressionParameter(ParameterBase):

    TYPE = ParameterType.OBJECT_REFERENCE

    def __init__(self, expression: dict):
        super().__init__()
        self.expression = expression


class PromptParameter(ParameterBase):

    TYPE = ParameterType.PROMPT

    def __init__(self, prompt: "ObjectReference"):
        super().__init__()
        self.prompt = prompt

    _FROM_DICT_MAP = {"prompt": ObjectReference.from_dict}


class DynamicDateTimeParameter(ParameterBase):

    TYPE = ParameterType.DYNAMIC_DATE_TIME

    def __init__(self, dynamic_date_time: dict):
        super().__init__()
        # TODO think about some object for dynamic_date_stucture
        self.dynamic_date_time = dynamic_date_time


class ConstantArrayParameter(ParameterBase):

    TYPE = ParameterType.ARRAY

    def __init__(self, type: "ConstantType", values: List[str]):
        super().__init__()
        self.constants_type = type
        self.constants = values

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source["type"] = source.get('constants_type')
        source["values"] = source.get('constants')
        return super().from_dict(source, connection)



from typing import List, Optional, TYPE_CHECKING, Union

from mstrio.access_and_security.security_filter import (AttributeRef, ObjectInformation,
                                                        Qualification)
from mstrio.api import changesets, security_filters
from mstrio.api.security_filters import ShowExpressionAs, UpdateOperator
from mstrio.object_management import full_search
from mstrio.users_and_groups import User, UserGroup
from mstrio.utils import helper
from mstrio.utils.entity import Entity, ObjectSubTypes, ObjectTypes, DeleteMixin

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.users_and_groups import UserOrGroup


def list_security_filters(connection: "Connection", to_dictionary: bool = False,
                          limit: Optional[int] = None, user: Optional[Union[str, "User"]] = None,
                          user_group: Optional[Union[str, "UserGroup"]] = None,
                          **filters) -> Union[List["SecurityFilter"], List[dict]]:
    """Get list of security filter objects or security filter dicts for the
    project specified in the connection object. It can be filtered by user or
    user group.

    Note: It is not possible to provide both `user` and `user_group` parameter.
        When both arguments are provided error is raised.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            `SecurityFilter` objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        user (str or object, optional): Id of user or `User` object used to
            filter security filters
        user_group (str or object, optional): Id of user group or `UserGroup`
            object used to filter security filters
        **filters: Available filter parameters: ['id', 'name', 'description',
            'type', 'subtype', 'date_created', 'date_modified', 'version',
            'acg', 'icon_path', 'owner']

    Returns:
        list of security filter objects or list of security filter dicts.
    """
    connection._validate_project_selected()
    if user and user_group:
        helper.exception_handler(
            "You cannot filter by both `user` and `user_group` at the same time.")

    if user:
        user = User(connection, id=user) if isinstance(user, str) else user
        # filter security filters by user for project specified in `connection`
        objects_ = user.list_security_filters(connection.project_id,
                                              to_dictionary=True).get(connection.project_name, [])
    elif user_group:
        user_group = UserGroup(connection, id=user_group) if isinstance(user_group,
                                                                        str) else user_group
        # filter security filters by the user group for project specified in
        # `connection`
        objects_ = user_group.list_security_filters(connection.project_id, to_dictionary=True).get(
            connection.project_name, [])
    else:
        objects_ = full_search(connection, object_types=ObjectTypes.SECURITY_FILTER,
                               project=connection.project_id, limit=limit, **filters)
    if to_dictionary:
        return objects_
    return [SecurityFilter.from_dict(obj_, connection) for obj_ in objects_]


class SecurityFilter(Entity, DeleteMixin):

    _OBJECT_TYPE = ObjectTypes.SECURITY_FILTER
    _OBJECT_SUBTYPE = ObjectSubTypes.MD_SECURITY_FILTER.value
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    def __init__(self, connection: "Connection", id: str, name: Optional[str] = None):
        """Initialize security filter object by its identifier.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify security filter.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing security filter containing
                the required data.
            name (str, optional): name of security filter.
        """
        super().__init__(connection, id, name=name)
        connection._validate_project_selected()

    def _get_definition(self):
        """Get the definition of security filter."""
        res = security_filters.read_security_filter(
            self.connection, self.id, self.connection.project_id,
            show_expression_as=ShowExpressionAs.TREE.value).json()
        self._information = ObjectInformation.from_dict(res['information'])
        self._qualification = Qualification.from_dict(res['qualification'])
        self._top_level = [AttributeRef.from_dict(level) for level in res.get('topLevel', [])]
        self._bottom_level = [
            AttributeRef.from_dict(level) for level in res.get('bottomLevel', [])
        ]
        self.__definition_retrieved = True

    def _get_members(self):
        """Get the users and user groups that the specified security filter is
        applied to."""
        res = security_filters.get_security_filter_members(
            self.connection,
            self.id,
            self.connection.project_id,
        ).json()
        self._members = []
        for obj in res['users']:
            if obj['subtype'] == ObjectSubTypes.USER.value:
                self._members.append(User.from_dict(obj, self.connection))
            if obj['subtype'] == ObjectSubTypes.USER_GROUP.value:
                self._members.append(UserGroup.from_dict(obj, self.connection))
        self.__members_retrieved = True

    def _init_variables(self, **kwargs):
        """Init all properties when creating security filter from
        a dictionary."""
        super()._init_variables(**kwargs)
        # properties `information`, `qualification`, `top_level` or
        # `bottom_level` can be lazily fetched
        information = kwargs.get("information")
        self._information = None if information is None else ObjectInformation.from_dict(
            information)
        qualification = kwargs.get("qualification")
        self._qualification = None if qualification is None else Qualification.from_dict(
            qualification)
        top_level = kwargs.get("top_level")
        self._top_level = None if top_level is None else [
            AttributeRef.from_dict(level) for level in top_level
        ]
        bottom_level = kwargs.get("bottom_level")
        self._bottom_level = None if top_level is None else [
            AttributeRef.from_dict(level) for level in bottom_level
        ]
        self.__definition_retrieved = self._information is not None and self._qualification is not\
            None and self._top_level is not None and self._bottom_level is not None
        # property `members` can be lazily fetched
        self._members = kwargs.get("members")
        self.__members_retrieved = self._members is not None

    def fetch(self, attr: Optional[str] = None, fetch_definition: bool = True,
              fetch_members: bool = True) -> None:
        """Fetch the latest security filter state from the I-Server.
        Additionally fetch its definition and members.

        Note:
            This method can overwrite local changes made to the object.

        Args:
            attr (str, optional): Attribute name to be fetched.
            fetch_definition (bool, optional): flag which indicates whether
                definition of security filter is fetched. By defualt `True`.
            fetch_members (bool, optional): flag which indicates whether
                members of security filter is fetched. By defualt `True`.

        Raises:
            ValueError: if `attr` cannot be fetched.
        """
        super().fetch(attr)
        if fetch_definition:
            self._get_definition()
        if fetch_members:
            self._get_members()

    @staticmethod
    def __prepare_body(qualification: dict, id: str = None, name: str = None,
                       folder_id: str = None, description: str = None, top_level: List[dict] = [],
                       bottom_level: List[dict] = []) -> dict:
        information = ObjectInformation(object_id=id, sub_type="md_security_filter", name=name,
                                        description=description, destination_folder_id=folder_id)

        return helper.delete_none_values({
            "information": information.to_dict(),
            "qualification": qualification,
            "topLevel": top_level,
            "bottomLevel": bottom_level
        })

    @staticmethod
    def __create_update_helper(func, connection: "Connection",
                               qualification: Union["Qualification", dict], name: str = None,
                               description: str = None,
                               top_level: Union[List[dict], List["AttributeRef"]] = [],
                               bottom_level: Union[List[dict], List["AttributeRef"]] = [],
                               id: str = None, folder_id: str = None):
        changeset_id = changesets.create_changeset(connection).json()['id']
        qualification = qualification.to_dict() if isinstance(qualification,
                                                              Qualification) else qualification
        top_level_dicts = []
        for level in top_level:
            if isinstance(level, AttributeRef):
                top_level_dicts.append(level.to_dict())
            else:
                top_level_dicts.append(level)
        bottom_level_dicts = []
        for level in bottom_level:
            if isinstance(level, AttributeRef):
                bottom_level_dicts.append(level.to_dict())
            else:
                bottom_level_dicts.append(level)

        res = func(
            connection=connection, changeset_id=changeset_id, throw_error=False, id=id,
            body=SecurityFilter.__prepare_body(qualification, id, name, folder_id, description,
                                               top_level_dicts, bottom_level_dicts))
        # there might be a case when response below is correct but commiting
        # changes is incorrect
        res_commit = changesets.commit_changeset_changes(connection, changeset_id,
                                                         throw_error=False)

        changesets.delete_changeset(connection, changeset_id)
        if not res_commit.ok:
            return res_commit
        return res

    @classmethod
    def create(cls, connection: "Connection", qualification: Union["Qualification", dict],
               name: str, folder_id: str, description: str = None,
               top_level: Union[List[dict], List["AttributeRef"]] = [],
               bottom_level: Union[List[dict], List["AttributeRef"]] = []) -> "SecurityFilter":
        """Create a security filter by providing its qualification, name and
        id of folder in which it will be created. Optionally it is possible to
        set description, top level and bottom level of security filter.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            qualification (object or dict): new security filter definition
                written as an expression tree over predicate nodes. It can be
                provided as `Qualification` object or dictionary.
            name (str): name of new security filter
            folder_id (str): id of folder in which new security filter will be
                created
            description (str): optional description of new security filter
            top_level (list of objects or list of dicts): the top level
                attribute list.
            bottom_level (list of object or list of dicts): the bottom level
                attribute list.
        """
        res = SecurityFilter.__create_update_helper(security_filters.create_security_filter,
                                                    connection, qualification, name, description,
                                                    top_level, bottom_level, folder_id=folder_id)
        if res.ok:
            res = res.json()
            return SecurityFilter.from_dict(
                {
                    "id": res['information']['objectId'],
                    **res['information'],
                    **res
                }, connection)
        else:
            return None

    def alter(self, qualification: Union["Qualification",
                                         dict] = None, top_level: Union[List[dict],
                                                                        List["AttributeRef"]] = [],
              bottom_level: Union[List[dict], List["AttributeRef"]] = [], name: str = None,
              description: str = None, folder_id: str = None):
        """Alter the security filter object in the metadata. It is possible to
        change its qualification, top level, bottom level, name, description
        or destination folder.

        Args:
            qualification (object or dict): optional new security filter
                definition written as an expression tree over predicate nodes.
                It can be provided as `Qualification` object or dictionary.
            top_level (list of objects or list of dicts): the top level
                attribute list.
            bottom_level (list of objects or list of dicts): the bottom level
                attribute list.
            name (str): optional new name of security filter
                top_level (list of object): the top level attribute list.
            description (str): optional new description of security filter
            folder_id (str): id of folder to which security filter will be
                moved
        """
        qualification = self.qualification if qualification is None else qualification
        name = self.name if name is None else name
        description = self.description if description is None else description
        folder_id = self.information.destination_folder_id if folder_id is None else folder_id
        res = self.__create_update_helper(security_filters.update_security_filter, self.connection,
                                          qualification, name, description, top_level,
                                          bottom_level, self.id, folder_id)
        if res.ok:
            res = res.json()
            members = self._members
            members_retrieved = self.__members_retrieved
            self._init_variables(**res.get('information'), id=res.get('information',
                                                                      {}).get('objectId'),
                                 connection=self.connection)
            self._members = members
            self.__members_retrieved = members_retrieved
            self._information = res.get('information')
            self._qualification = res.get('qualification')
            self._top_level = res.get('top_level')
            self._bottom_level = res.get('bottom_level')
            self.__definition_retrieved = self._information is not None and self._qualification is\
                not None and self._top_level is not None and self._bottom_level is not None

    @staticmethod
    def __retrieve_ids_from_list(
        objects: Union[str, "User", "UserGroup", List[Union[str, "User",
                                                            "UserGroup"]]]) -> List[str]:
        """Parsing a list which can contain at the same time User object(s),
        UserGroup object(s), id(s) to a list with id(s)."""
        objects = objects if isinstance(objects, list) else [objects]
        ids = []
        for obj in objects:
            ids.append(obj if isinstance(obj, str) else obj.id)
        return ids

    def __update_members(self, op: UpdateOperator,
                         users_and_groups: Union["UserOrGroup", List["UserOrGroup"]] = []) -> bool:
        """Update members of security filter."""
        users_or_groups = self.__retrieve_ids_from_list(users_and_groups)
        body = {"operationList": [{"op": op.value, "path": "/members", "value": users_or_groups}]}
        res = security_filters.update_security_filter_members(self.connection, self.id, body,
                                                              self.connection.project_id,
                                                              throw_error=False)
        if res.ok:
            self._get_members()
        return res.ok

    def apply(self, users_and_groups: Union["UserOrGroup", List["UserOrGroup"]] = []) -> bool:
        """Apply security filter to user(s) and/or user group(s).

        Args:
            users_and_groups (string, object, list of strings or objects): list
            or a single element specifying to which users and user groups
            security filter will be applied. It is possible to provide the whole
            User or UserGroup object(s) or just id(s) of those objects. When
            providing a list it can contain User object(s), UserGroup object(s),
            id(s) at the same time.

        Returns:
            True when applying was successful. False otherwise.
        """

        return self.__update_members(UpdateOperator.APPLY, users_and_groups)

    def revoke(self, users_and_groups: Union["UserOrGroup", List["UserOrGroup"]] = []) -> bool:
        """Revoke security filter from user(s) and/or user group(s).

        Args:
            users_and_groups (string, object, list of strings or objects): list
            or a single element specifying from which users and user groups
            security filter will be revoked. It is possible to provide the whole
            User or UserGroup object(s) or just id(s) of those objects. When
            providing a list it can contain User object(s), UserGroup object(s),
            id(s) at the same time

        Returns:
            True when revoking was successful. False otherwise.

        """
        return self.__update_members(UpdateOperator.REVOKE, users_and_groups)

    @property
    def information(self):
        """All of the type-neutral information stored in the metadata about
        a security filter."""
        if not self.__definition_retrieved:
            self._get_definition()
        return self._information

    @property
    def qualification(self):
        """The security filter definition written as an expression tree over
        predicate nodes."""
        if not self.__definition_retrieved:
            self._get_definition()
        return self._qualification

    @property
    def top_level(self):
        """The top level attribute list."""
        if not self.__definition_retrieved:
            self._get_definition()
        return self._top_level

    @property
    def bottom_level(self):
        """The bottom level attribute list."""
        if not self.__definition_retrieved:
            self._get_definition()
        return self._bottom_level

    @property
    def members(self):
        """List of users and user groups that security filter is applied to."""
        if not self.__members_retrieved:
            self._get_members()
        return self._members



from typing import Optional, Union, List

from pandas import DataFrame

from mstrio.connection import Connection
from mstrio.server.environment import Environment
from mstrio.users_and_groups import User, list_users, UserGroup, UserOrGroup
from mstrio.api import documents, library
from mstrio.utils import helper
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import Entity, VldbMixin, ObjectTypes


def list_documents(connection: Connection, name: Optional[str] = None, to_dictionary: bool = False,
                   to_dataframe: bool = False, limit: Optional[int] = None, **filters):
    """Get all Documents available in the project specified within the
    `connection` object.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        name: exact name of the document to list
        to_dictionary(bool, optional): if True, return Documents as
            list of dicts
        to_dataframe(bool, optional): if True, return Documents as
            pandas DataFrame
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of documents.
    """
    # TODO: consider adding Connection.project_selected attr/method
    if connection.project_id is None:
        raise ValueError(("Please log into a specific project to load documents within it. "
                          "To load all documents across the whole environment use "
                          f"{list_documents_across_projects.__name__} function."))
    return Document._list_all(connection, to_dictionary=to_dictionary, name=name, limit=limit,
                              to_dataframe=to_dataframe, **filters)


def list_documents_across_projects(connection: Connection, name: Optional[str] = None,
                                   to_dictionary: bool = False, to_dataframe: bool = False,
                                   limit: Optional[int] = None, **filters):
    """Get all Documents stored on the server.

    Optionaly use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        name: exact names of the documents to list
        to_dictionary(bool, optional): if True, return Documents as
            list of dicts
        to_dataframe(bool, optional): if True, return Documents as
            pandas DataFrame
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of documents.
    """
    project_id_before = connection.project_id
    env = Environment(connection)
    projects = env.list_projects()
    output = []
    for project in projects:
        connection.select_project(project_id=project.id)
        output.extend(
            Document._list_all(connection, to_dictionary=to_dictionary, name=name, limit=limit,
                               to_dataframe=to_dataframe, **filters))
        output = list(set(output))
    connection.select_project(project_id=project_id_before)
    return output


class Document(Entity, VldbMixin):
    _OBJECT_TYPE = ObjectTypes.DOCUMENT_DEFINITION
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP, 'owner': User.from_dict,
        'certified_info': CertifiedInfo.from_dict
    }

    def __init__(self, connection: Connection, name: Optional[str] = None,
                 id: Optional[str] = None):
        """Initialize Document object by passing name or id.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            name: name of Document
            id: ID of Document
        """
        if id is None and name is None:
            raise ValueError("Please specify either 'name' or 'id' parameter in the constructor.")
        if id is None:
            documents = self._list_all(connection=connection, to_dictionary=True, name=name)
            if documents:
                id = documents[0]['id']
            else:
                msg = (f"There is no {self.__class__.__name__} associated with the given "
                       f"name: '{name}'")
                raise ValueError(msg)

        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._instance_id = ""
        self._recipients = []

    def alter(self, name: Optional[str] = None, description: Optional[str] = None):
        """Alter Document name or/and description.

        Args:
            name: new name of the Document
            description: new description of the Document
        """
        func = self.alter
        args = func.__code__.co_varnames[:func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def publish(self, recipients: Optional[Union[UserOrGroup, List[UserOrGroup]]] = None):
        """Publish the document for authenticated user. If `recipients`
        parameter is specified publishes the document for the given users.

        Args:
            recipients(list): list of users or user groups to publish the
                document to (can be a list of IDs or a list of User and
                UserGroup elements)
        """
        if not isinstance(recipients, list) and recipients is not None:
            recipients = [recipients]

        if recipients is None:
            recipients = [self.connection.user_id]
        elif all([isinstance(el, User) for el in recipients]):
            recipients = [recipient.id for recipient in recipients]
        elif all([isinstance(el, UserGroup) for el in recipients]):
            users = [user for group in recipients for user in group.members]
            recipients = [user["id"] for user in users]
        elif any([not isinstance(el, str) for el in recipients]):
            raise ValueError('Please provide either list of User, UserGroup or str elements.')
        body = {'id': self.id, 'recipients': recipients}
        self._instance_id = ''
        library.publish_document(self.connection, body)

    def unpublish(self, recipients: Optional[Union[UserOrGroup, List[UserOrGroup]]] = None):
        """Unpublish the document for all users it was previously published to.
        If `recipients` parameter is specified unpublishes the document for the
        given users.

        Args:
            recipients(list): list of users or user groups to publish the
                document to (can be a list of IDs or a list of User and
                UserGroup elements).
        """

        if recipients is None:
            library.unpublish_document(self.connection, id=self.id)
        else:
            if not isinstance(recipients, list):
                recipients = [recipients]
            if all([isinstance(el, User) for el in recipients]):
                recipients = [recipient.id for recipient in recipients]
            elif all([isinstance(el, UserGroup) for el in recipients]):
                users = [user for group in recipients for user in group.members]
                recipients = [user["id"] for user in users]
            elif any([not isinstance(el, str) for el in recipients]):
                raise ValueError(
                    'Please provide either list User and UserGroup elements or str elements.')
            for user_id in recipients:
                library.unpublish_document_for_user(self.connection, document_id=self.id,
                                                    user_id=user_id)

    def share_to(self, users: Union[UserOrGroup, List[UserOrGroup]]):
        """Shares the document to the listed users' libraries.

        Args:
            users(list): list of users or user groups to publish the
                document to (can be a list of IDs or a list of User and
                UserGroup elements).
        """
        self.publish(users)

    @classmethod
    def _list_all(cls, connection: Connection, name: Optional[str] = None,
                  to_dictionary: bool = False, to_dataframe: bool = False,
                  limit: Optional[int] = None,
                  **filters) -> Union[List["Document"], List[dict], DataFrame]:
        msg = "Error retrieving documents from the environment."
        if to_dictionary and to_dataframe:
            helper.exception_handler(
                "Please select either `to_dictionary=True` or `to_dataframe=True`, but not both.",
                ValueError)
        objects = helper.fetch_objects_async(connection, api=documents.get_documents,
                                             async_api=documents.get_documents_async,
                                             dict_unpack_value='result', limit=limit,
                                             chunk_size=1000, error_msg=msg, filters=filters,
                                             search_term=name)
        if to_dictionary:
            return objects
        elif to_dataframe:
            return DataFrame(objects)
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    def get_connected_cubes(self):
        """Lists cubes used by this document."""
        cubes = documents.get_cubes_used_by_document(self.connection, self.id).json()
        ret_cubes = [helper.choose_cube(self.connection, cube) for cube in cubes]
        return [tmp for tmp in ret_cubes if tmp]  # remove `None` values

    @property
    def instance_id(self):
        if self._instance_id == '':
            body = {"resolveOnly": True, "persistViewState": True}
            response = documents.create_new_document_instance(connection=self.connection,
                                                              document_id=self.id, body=body)
            self._instance_id = response.json()['mid']
        return self._instance_id

    @property
    def recipients(self):
        response = library.get_document(connection=self.connection,
                                        id=self.id).json()['recipients']
        if response:
            self._recipients = list_users(connection=self.connection,
                                          id=[r['id'] for r in response])
        else:
            self._recipients = []
        return self._recipients
